#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <ctype.h>
#include <dlfcn.h>
#include <errno.h>
#include <time.h>

#include "eiwic.h"
#include "settings.h"
#include "plugins.h"



/*--- settings functions ---*/

SETTING *set_find(char *label)
{
	SETTING *set;
	FOR_EACH_DATA(&e_global->settings, set)
	{
		if (strcasecmp(set->label, label) == 0)
		{
			return set;
		}
	}
	END_DATA;

	return NULL;
}

char *set_get_string(char *label)
{
	SETTING *s = set_find(label);

	if (s)
	switch(s->type)
	{
	case SET_STRING:
	case SET_LONGSTRING:
		return s->data.d_string;
	}

	return "";
}


long set_get_int(char *label)
{
	SETTING *s = set_find(label);

	if (s)
	switch(s->type)
	{
	case SET_INT:
		return s->data.d_int;
	}

	return 0;
}

float set_get_float(char *label)
{
	SETTING *s = set_find(label);

	if (s)
	switch(s->type)
	{
	case SET_FLOAT:
		return s->data.d_float;
	}

	return 0;
}

int set_is(char *label)
{
	SETTING *s = set_find(label);

	if (s)
	switch(s->type)
	{
	case SET_BOOL:
		return s->data.d_int;
	case SET_INT:
		return ((s->data.d_int > 0)?1:0);
	case SET_STRING:
	case SET_LONGSTRING:
		return ((strlen(s->data.d_string) > 0)?1:0);
	}

	return 0;
}

void set_set(SETTING *s, ...)
{
	va_list ap;
	char *x;

	va_start(ap, s);	
	switch (s->type)
	{
	case SET_BOOL:
	case SET_INT:
			s->data.d_int = va_arg(ap, long);
#ifdef VV_DEBUG
			log(LOG_DEBUG, "Setting '%s' changed to '%i'",
					s->label, s->data.d_int);
#endif
		break;
	case SET_FLOAT:
			s->data.d_float = va_arg(ap, double);
#ifdef VV_DEBUG
			log(LOG_DEBUG, "Setting '%s' changed to '%2.2f'",
					s->label, s->data.d_float);
#endif
		break;
	case SET_STRING:
	case SET_LONGSTRING:
			x = va_arg(ap, char *);
			if (s->type == SET_STRING)
			{
				s->data.d_string = realloc(s->data.d_string, MIN(strlen(x), 128) + 1);
				strncpy(s->data.d_string, x, MIN(strlen(x), 128));
				s->data.d_string[MIN(strlen(x), 128)] = 0;
			}
			else
			{
				s->data.d_string = realloc(s->data.d_string, strlen(x) + 1);
				strcpy(s->data.d_string, x);
			}
#ifdef VV_DEBUG
			log(LOG_DEBUG, "Setting '%s' changed to '%s'",
					s->label, s->data.d_string);
#endif
		break;
	}
	va_end(ap);
}

SETTING *set_add(int type, char *label, ...)
{
	va_list ap;
	SETTING *s;

	if (type < SET_BOOL || type > SET_LONGSTRING)
	{
		log(LOG_DEBUG, "Couldn't add setting: Unknown type: %i", type);
		return NULL;
	}

	s = malloc(sizeof(SETTING));
	strncpy(s->label, label, 63);
	s->type = type;
	s->data.d_string = NULL;
	s->data.d_int = 0;
	s->data.d_float = 0.0;

	va_start(ap, label);
	set_set(s, va_arg(ap, void*));
	va_end(ap);
	
	dlist_ins_next(&e_global->settings, dlist_tail(&e_global->settings), s);
	
	log(LOG_VERBOSE_DEBUG, "Setting '%s' (type %i) has been added.", s->label, s->type);

	return s;
}

void set_remove(SETTING *set)
{
	return;
}

void set_destroy(void *s)
{
	SETTING *set = s;

	if (set->type == SET_STRING ||
			set->type == SET_LONGSTRING)
	{
		free(set->data.d_string);
	}

	free(s);
}

/*--- end of settings functions ---*/



void log(int verbose_level, char *line, ...)
{
	va_list m;
	FILE *out;

	if (e_global->var_log == 0)
	{
		if (verbose_level <= LOG_WARNING)
			out = stderr;
		else
			return;
	}
	else
		out = e_global->var_logfile;
	
	if (e_global->var_verbose <= verbose_level)
		return;
	
	if (plugin)
	{
		fprintf(out, "%s: ", plugin->name);
	}
	else
	{
		fprintf(out, "eiwic: ");
	}
	
	switch (verbose_level)
	{
		case LOG_ERROR:
				fprintf(out, "Error: ");
			break;
		case LOG_WARNING:
				fprintf(out, "Warning: ");
			break;
		case LOG_STATUS:
		case LOG_DEBUG:
		case LOG_VERBOSE_DEBUG:
		case LOG_MISC:
			break;
	}
	
	va_start(m, line);
	vfprintf(out, line, m);
	va_end(m);
	
	if (errno && verbose_level <= LOG_WARNING)
	{
		fprintf(out, " (%s)", strerror(errno));
	}
	
	fprintf(out, "\n");
	fflush(out);
}

void plog(PLUGIN *plugin, int verbose_level, char *line, ...)
{

	va_list m;
	FILE *out;
	
	if (e_global->var_log == 0)
	{
		if (verbose_level <= LOG_WARNING)
			out = stderr;
		else
			return;
	}
	else
		out = e_global->var_logfile;
	
	if (e_global->var_verbose <= verbose_level)
		return;
	
	if (plugin)
	{
		fprintf(out, "%s: ", plugin->name);
	}
	else
	{
		fprintf(out, "eiwic: ");
	}
	
	switch (verbose_level)
	{
		case LOG_ERROR:
				fprintf(out, "Error: ");
			break;
		case LOG_WARNING:
				fprintf(out, "Warning: ");
			break;
		case LOG_STATUS:
		case LOG_DEBUG:
		case LOG_VERBOSE_DEBUG:
		case LOG_MISC:
			break;
	}
	
	va_start(m, line);
	vfprintf(out, line, m);
	va_end(m);
	
	if (errno && verbose_level <= LOG_WARNING)
	{
		fprintf(out, " (%s)", strerror(errno));
	}
	
	fprintf(out, "\n");
	fflush(out);
}




/*--- channel functions ---*/

CHANNEL *channel_add(char *name)
{
	CHANNEL *chan;

	chan = malloc(sizeof(CHANNEL));
	bzero(chan, sizeof(CHANNEL));

	strncpy(chan->name, name, sizeof(chan->name) - 1);

	dlist_ins_next(&e_global->channels, dlist_tail(&e_global->channels), chan);
	log(LOG_VERBOSE_DEBUG, "Channel %s has been added.", chan->name);

	return chan;
}

void channel_update(void)
{
	DListElmt *el;
	
	FOR_EACH(&e_global->channels, el)
	{
		channel_join(el->data);
	}
}

void channel_remove(CHANNEL *chan)
{
	return;
}

void channel_destroy(void *data)
{
	CHANNEL *chan = data;

	free(chan);
}

void channel_join(void *which)
{
	CHANNEL *chan = which;
	
	if (chan == NULL)
		return;
	
	log(LOG_VERBOSE_DEBUG, "Attempting to join %s", chan->name);
	plug_sendf("JOIN %s\r\n", chan->name);
}

CHANNEL *channel_find(char *name)
{
	DListElmt *el;

	FOR_EACH(&e_global->channels, el)
		if (strcasecmp(((CHANNEL *)el->data)->name, name) == 0)
			return (CHANNEL *)el->data;
		
	return NULL;
}

/*--- end of channel functions ---*/











/*--- plugin functions ---*/

PLUGIN *plug_load(char *filename)
{
	struct eiwic_plugin *n;
	void *handle;
	char fn[1024],i=0,*modpath;
	char *(*ep_getname)(), *name;

	log(LOG_DEBUG, "plug_load: Prepare loading of %s...", filename);

	modpath = set_get_string("module_path");
	do {
		switch (i)
		{
		case 0: sprintf(fn, "%s%s.so",modpath,filename); break;
		case 1: sprintf(fn, "%s%s",modpath,filename); break;
		case 2: sprintf(fn, "%s%s.dll",modpath,filename); break;
		case 3:	sprintf(fn, "%s", filename); break;
		case 4:	log(LOG_ERROR, "Failed to load module %s", filename); return NULL;
		}

		log(LOG_DEBUG, "plug_load: Trying %s", fn);

#ifdef __OpenBSD__
		handle = dlopen(fn, DL_LAZY);
#else
		handle = dlopen(fn, RTLD_NOW);
#endif

		i++;

	} while (handle == NULL); 

#ifdef __OpenBSD__
	ep_getname = dlsym(handle, "_ep_getname");
#else
	ep_getname = dlsym(handle, "ep_getname");
#endif
	
	if (ep_getname == NULL)
	{	
		log(LOG_ERROR, "Failed to find routine 'ep_getname' from module %s", filename);
		dlclose(handle);
		return NULL;
	}

	if (
#ifdef __OpenBSD__
			dlsym(handle, "_ep_init")
#else
			dlsym(handle, "ep_init")
#endif
			== NULL)
	{
		log(LOG_ERROR, "Failed to find routine 'ep_init' from module %s", filename);
		dlclose(handle);
		return NULL;
	}

	name = ep_getname();
	
	if (name == NULL)
	{	
		log(LOG_ERROR, "Failed to init module %s through routine 'ep_getname'", filename);
		dlclose(handle);
		return NULL;
	}
	else
	{
		n = malloc(sizeof(PLUGIN));
		bzero(n, sizeof(PLUGIN));
			
		n->handle = handle;
		n->modlink = e_global->modlink;
				
		strcpy(n->name, name);
		strcpy(n->filename, fn);

#ifdef __OpenBSD__
		n->ep_init = dlsym(handle, "_ep_init");
		n->ep_main = dlsym(handle, "_ep_main");
		n->ep_deinit = dlsym(handle, "_ep_deinit");
		n->ep_parse = dlsym(handle, "_ep_parse");
		n->ep_trigger = dlsym(handle, "_ep_trigger");
#else
		n->ep_init = dlsym(handle, "ep_init");
		n->ep_main = dlsym(handle, "ep_main");
		n->ep_deinit = dlsym(handle, "ep_deinit");
		n->ep_parse = dlsym(handle, "ep_parse");
		n->ep_trigger = dlsym(handle, "ep_trigger");
#endif

		plug_add(n);

		e_global->mod_need_init = 1;
	}
	
	return n;
}

int plug_unload(PLUGIN *ep)
{
	plug_remove(ep);
	plug_destroy(ep);

	return 1;
}

void plug_add(PLUGIN *ep)
{
	dlist_ins_next(&e_global->plugins, dlist_tail(&e_global->plugins), ep);
	log(LOG_VERBOSE_DEBUG, "Plugin %s has been added.", ep->name);
#ifdef VV_DEBUG
	log(LOG_VERBOSE_DEBUG, "--> name: %s; ep_init: %i; handle: %i; filename: %s;", ep->name, ep->ep_init, ep->handle,  ep->filename);
#endif
	
}

void plug_remove(PLUGIN *ep)
{
	dlist_remove(&e_global->plugins, dlist_find(&e_global->plugins, ep), NULL);	
}

void plug_destroy(void *data)
{
	PLUGIN *ep = data;
	
	dlclose(ep->handle);
	free(ep);
}

/*--- end of plugins functions ---*/






/*--- trigger functions ---*/

TRIGGER *plug_trigger_reg(char *trigger, u_int flags, e_trigger_f func)
{
	TRIGGER *ept;
		
	ept = malloc(sizeof(struct eiwic_plugin_trigger));
	bzero(ept, sizeof(struct eiwic_plugin_trigger));
		
	strcpy(ept->trigger, trigger);
	ept->func = func;
	ept->plugin = plugin;
	ept->flags = flags;

	log(LOG_DEBUG, "Added trigger: %s (plugin %p)", trigger, plugin);
	
	dlist_ins_next(&e_global->triggers, e_global->triggers.tail, ept);

	return ept;
}

int plug_trigger_unreg(TRIGGER *trigger)
{
	dlist_remove(&e_global->triggers, dlist_find(&e_global->triggers, trigger), NULL);
	plug_trigger_destroy(trigger);
	
	return 1;
}

void plug_trigger_destroy(void *data)
{
	TRIGGER *ept = data;
	free(ept);
}

/*--- end of trigger functions ---*/








/*--- timer functions ---*/

TIMER *plug_timer_reg(u_long ut, e_timer_f func, void *data)
{
	TIMER *ept;
	
	ept = malloc(sizeof(TIMER));
	bzero(ept, sizeof(TIMER));
		
	ept->time = ut;
	ept->func = func;
	ept->plugin = plugin;
	ept->data = data;
	
	dlist_ins_next(&e_global->timers, e_global->timers.tail, ept);

	return ept;
}

int plug_timer_unreg(TIMER *timer)
{
	dlist_remove(&e_global->timers, dlist_find(&e_global->timers, timer), NULL);	
	plug_timer_destroy(timer);
	return 0;
}

void plug_timer_destroy(void *data)
{
	TIMER *timer = data;
	free(timer);
}

int plug_timers(void)
{
	TIMER *t;
	PLUGIN *p;
	e_timer_f on_timer;
	void *data;
	DListElmt *el;
	
start:
	FOR_EACH(&e_global->timers, el)
	{
		t = el->data;
		
		if (t->time < time(NULL))
		{
			data = t->data;
			on_timer = t->func;
			p = t->plugin;
			plug_timer_unreg(t);
			
			e_global->jump_back = 1;
			if (sigsetjmp(e_global->jmp_past, 1) == 0)
			{
				plugin = p;
				on_timer(data);
				plugin = NULL;
			}
			else
			{
				e_global->jump_back = 0;
				SEGWARN(p, "e_timer_f");
			}
			e_global->jump_back = 0;
			
			goto start;
		}
	}
	
	return 1;
}

/*--- end of timer functions ---*/

void strtolower(char *p)
{
	for(;*p;++p) *p = tolower(*p);
}

void strtoupper(char *p)
{
	for(;*p;++p) *p = toupper(*p);
}

int findip(char *hostname, struct in_addr *addr)
{
	struct hostent *hent;

	hent = gethostbyname(hostname);
	if (hent == NULL)
		return -1;
	
	bzero(addr, sizeof(struct in_addr));
	memcpy(addr, hent->h_addr, hent->h_length);
	
	if ((addr->s_addr == INADDR_ANY) ||
		(addr->s_addr == INADDR_NONE))
		addr->s_addr = inet_addr(hostname);
	
	if ((addr->s_addr == INADDR_ANY) ||
		(addr->s_addr == INADDR_NONE))
		return -1;
	else
		return 1;
}

#ifdef IPV6
int findip6(char *hostname, struct in6_addr *addr)
{
	struct hostent *hent;

	hent = gethostbyname2(hostname, AF_INET6);
	if (hent == NULL)
		return -1;
	
	bzero(addr, sizeof(struct in6_addr));
	memcpy(addr, hent->h_addr, hent->h_length);
	
	return 1;
}

char *inet6_ntoa(struct in6_addr in6)
{
	static char static_buffer[sizeof("1234:5678:9abc:def0:1234:5678:9abc:def0")];
	char *buffer;
	uint8_t *p;

	buffer = static_buffer;
	p = (uint8_t *)&in6;
	
	sprintf(buffer, "%02x%02x:%02x%02x:%02x%02x:%02x%02x:"
					"%02x%02x:%02x%02x:%02x%02x:%02x%02x",
					p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7],
					p[8], p[9], p[10], p[11], p[12], p[13], p[14], p[15]);
	
	return buffer;
};
#endif 
	
int plug_sendf(char *fmt, ...)
{
	va_list m;
	char buf[1024];

	if (fmt == NULL)
		return plug_send(NULL);
	
	va_start(m, fmt);
	vsprintf(buf, fmt, m);
	va_end(m);
	
	return plug_send(buf);
}

int plug_send(char *s)
{
	u_int count;

	if (e_global->var_noconnect == 1)
		return 1;
	
	if (s)
	{
		if (e_global->b_send)
		{
//			printf("Send buffer realloced to %i bytes..", strlen(e_global->b_send) + strlen(s) + 1);
			//fflush(stdout);
			e_global->b_send = realloc(e_global->b_send, strlen(e_global->b_send) + strlen(s) + 1);
			strcat(e_global->b_send, s);
		}
		else
		{
			//printf("Send buffer allocated, %i bytes..", strlen(s) + 1);
			//fflush(stdout);
			e_global->b_send = malloc(strlen(s) + 1);
			strcpy(e_global->b_send, s);			
		}
	}

	if (e_global->b_send == NULL) return 0;
		
	if (e_global->var_bytes >= 200)
		return 0;
	else	
		count = 200 - e_global->var_bytes;
		
	if (strlen(e_global->b_send) <= count)
	{
	/*	printf("Send buffer freed, sent %i bytes. %s\n", strlen(e_global->b_send), e_global->b_send);
		fflush(stdout);*/
	/*	log(LOG_DEBUG, "=> %s", e_global->b_send); */
		send(e_global->var_socket, e_global->b_send, strlen(e_global->b_send), 0);
		
		e_global->var_bytes += strlen(e_global->b_send);
		
		free(e_global->b_send);
		e_global->b_send = NULL;
	}
	else
	{
		/*printf("_Send buffer sent, %i bytes %i\n", count, e_global->b_send);
		fflush(stdout);*/
	/*	log(LOG_DEBUG, "=> [%d]%s", count, e_global->b_send); */
		send(e_global->var_socket, e_global->b_send, count, 0);
		
		e_global->var_bytes += count;
		
		memcpy(e_global->b_send, e_global->b_send + count, strlen(e_global->b_send) - count);
		e_global->b_send[strlen(e_global->b_send) - count] = 0;
		e_global->b_send = realloc(e_global->b_send, strlen(e_global->b_send) + 1);
	}
	
	return 1;
}
